# -*- coding: utf-8 -*-

from astrobox.space_field import SpaceField
from astrobox.core import Drone, Unit
from astrobox.utils import nearest_angle_distance
from demo.strategies import *
from robogame_engine.geometry import Vector
from robogame_engine.theme import theme


class DroneUnitWithStrategies(Drone):
    def __init__(self, **kwargs):
        super(DroneUnitWithStrategies, self).__init__(**kwargs)
        self.__strategies = []

    def append_strategy(self, strategy):
        if strategy.is_group_unique:
            for s in self.__strategies:
                if s.group == strategy.group:
                    self.__strategies.remove(s)
        self.__strategies.append(strategy)

    def clear_strategies(self):
        self.__strategies = []

    def is_strategy_finished(self):
        return len(self.__strategies) == 0

    def game_step(self):
        self.native_game_step()
        for s in self.__strategies:
            if s.is_finished:
                self.__strategies.remove(s)
                continue
            s.game_step()
            break

    @property
    def elerium_stocks(self):
        """возвращает все объекты мира из которых можно добывать ресурсы """
        return [es for es in self.scene.get_objects_by_type(Unit) if hasattr(es, 'cargo') and not es.is_alive]

    def native_game_step(self):
        """Позволяет обращаться к чистому обработчику из стратегий """
        super(DroneUnitWithStrategies, self).game_step()


class WorkerDrone(DroneUnitWithStrategies):
    counter_attrs = dict(size=22, position=(75, 135), color=(255, 255, 255))

    def __init__(self, **kwargs):
        super(WorkerDrone, self).__init__(**kwargs)
        self._elerium_stock = None

    @property
    def elerium_stock(self):
        return self._elerium_stock

    def set_elerium_stock(self, stock):
        self._elerium_stock = stock

    def on_born(self):
        super(WorkerDrone, self).on_born()
        self.append_strategy(StrategyHarvesting(unit=self))


class GreedyDrone(WorkerDrone):

    def __init__(self, **kwargs):
        super(GreedyDrone, self).__init__(**kwargs)

    def get_nearest_elerium_stock(self):
        elerium_stocks = [es for es in self.scene.asteroids if es.cargo.payload > 0]
        for drone in self.teammates:
            if drone.elerium_stock is not None and \
                    not drone.cargo.is_full and \
                    drone.elerium_stock in elerium_stocks:
                elerium_stocks.remove(drone.elerium_stock)

        if not elerium_stocks:
            return None
        # Берем наибольшее кол-во elerium-а, что сможем унести, из ближайшего
        elerium_stocks = sorted(elerium_stocks, key=lambda x: x.distance_to(self))
        nearest_stock = None
        max_elerium = 0
        for stock in elerium_stocks:
            if stock.cargo.payload >= self.cargo.free_space:
                return stock
            if stock.cargo.payload > max_elerium:
                nearest_stock = stock
                max_elerium = stock.cargo.payload
        if nearest_stock:
            return nearest_stock
        return random.choice(elerium_stocks)


class HunterDrone(GreedyDrone):
    _hunters = []

    def __init__(self, **kwargs):
        super(HunterDrone, self).__init__(**kwargs)
        self._hunting_strategy = None
        self._approach_strategy = None
        self._victim = None
        # self._no_victim_strategy = False
        self._victim_stamp = 0
        self._next_victim = None
        self.substrategy = None

    @property
    def victim(self):
        return self._victim

    def on_born(self):
        super(WorkerDrone, self).on_born()
        if self.have_gun:
            self._hunting_strategy = StrategyHunting.getTeamStrategy(self.team, self)
        else:
            self.append_strategy(StrategyHarvesting(unit=self))

    def on_stop(self):
        pass

    def get_nearest_elerium_stock(self):
        # Сперва сбор элериума с жертв
        elerium_stocks = [drone for drone in self.scene.drones
                          if not drone.is_alive and drone.cargo.payload > 0]
        for drone in self.teammates:
            if drone.elerium_stock is not None and \
                    not drone.cargo.is_full and \
                    drone.elerium_stock in elerium_stocks:
                elerium_stocks.remove(drone.elerium_stock)
        if elerium_stocks:
            elerium_stocks = sorted(elerium_stocks, key=lambda x: x.distance_to(self))
            return elerium_stocks[0]

        # Потом с астероидов
        elerium_stocks = [asteriod for asteriod in self.scene.asteroids if asteriod.cargo.payload > 0]
        for drone in self.teammates:
            if drone.elerium_stock is not None and \
                    not drone.cargo.is_full and \
                    drone.elerium_stock in elerium_stocks:
                elerium_stocks.remove(drone.elerium_stock)
        if not elerium_stocks:
            return None
        elerium_stocks = sorted(elerium_stocks, key=lambda x: x.distance_to(self))
        return elerium_stocks[0]

    def set_victim(self, victim):
        self._next_victim = None
        self._victim = victim
        self._victim_stamp = 0
        if not self.substrategy.is_finished:
            self.stop()
            self.state.stop()
        self.substrategy.reset()
        return victim.coord.copy()

    @property
    def is_unloading(self):
        return self.cargo.is_full or (self.substrategy is not None and
                                      self.substrategy.current_strategy_id == "approach&unload")

    def game_step(self):
        if not self.have_gun:
            super(HunterDrone, self).game_step()
            return

        self.native_game_step()
        if self._hunting_strategy is None:
            return
        self._hunting_strategy.game_step(self)
        if self.victim is not None:
            vector = Vector.from_points(self.coord, self.victim.coord,
                                        module=self.gun.shot_distance)
            if int(self.distance_to(self.victim)) < 1 or (
                    self.distance_to(self.victim) < vector.module
                    and abs(nearest_angle_distance(vector.direction, self.direction)) < 7
            ):
                self.gun.shot(self.victim)
        else:
            enemies = [enemy for enemy in self.scene.drones
                       if enemy.team != self.team and enemy.is_alive and
                       enemy.distance_to(self) < self.gun.shot_distance]
            enemie = sorted(enemies, key=lambda x: -x.cargo.payload)
            for enemy in enemies:
                vector = Vector.from_points(self.coord, enemy.coord)
                if abs(nearest_angle_distance(vector.direction, self.direction)) < 7:
                    self.gun.shot(enemy)
                    break
        pass


class RunnerDrone(DroneUnitWithStrategies):

    def any_asteroid(self):
        return random.choice(self.scene.asteroids)

    def on_born(self):
        self.append_strategy(StrategyApproach(unit=self, target_point=self.any_asteroid().coord, distance=0))

    def game_step(self):
        super(RunnerDrone, self).game_step()
        if self.is_strategy_finished():
            self.append_strategy(StrategyApproach(unit=self, target_point=self.any_asteroid().coord, distance=0))


class DestroyerDrone(DroneUnitWithStrategies):
    _hunters = []
    team = 'Destroyer'

    def __init__(self, **kwargs):
        super(DestroyerDrone, self).__init__(**kwargs)
        self._victim = None
        self._next_victim = None
        self._target_mship = None
        self._elerium_stock = None

    @property
    def elerium_stock(self):
        return self._elerium_stock

    def set_elerium_stock(self, stock):
        self._elerium_stock = stock

    def on_born(self):
        if self.have_gun:
            self.append_strategy(StrategyDestroyer(unit=self))
        else:
            self.append_strategy(StrategyHarvesting(unit=self))

    def game_step(self):
        super(DestroyerDrone, self).game_step()
        if self.have_gun:
            if self.is_strategy_finished():
                self.append_strategy(StrategyHarvesting(unit=self))


class CowardDrone(DestroyerDrone):

    def on_born(self):
        self.append_strategy(StrategyHarvesting(unit=self))


if __name__ == '__main__':
    space_field = SpaceField(
        name="Space war",
        speed=5,
        field=(1600, 800),
        asteroids_count=3,
        # can_fight=True,
        max_drones_at_team=12,
        # headless=True,
    )

    DRONES_COUNT = 5

    teamA = [WorkerDrone() for _ in range(DRONES_COUNT)]
    teamB = [GreedyDrone() for _ in range(DRONES_COUNT)]
    teamC = [HunterDrone() for _ in range(DRONES_COUNT)]
    teamD = [DestroyerDrone() for _ in range(DRONES_COUNT)]
    teamD.extend(CowardDrone() for _ in range(2))

    space_field.go()



# -*- coding: utf-8 -*-
import math
import random

from robogame_engine import GameObject
from robogame_engine.constants import ROTATE_TURNING
from robogame_engine.events import GameEvent

from .guns import PlasmaGun
from .cargo import Cargo, CargoTransition
from .theme import theme


class EventWakeUp(GameEvent):

    def handle(self, obj):
        obj.on_wake_up()


class Unit(GameObject):
    coord = None  # переопределяется в потомках
    radius = 0
    _cargo = None

    def __init__(self, payload=0, max_payload=1, **kwargs):
        super(Unit, self).__init__(**kwargs)
        self._move_target = None
        self._cargo = Cargo(self, payload=payload, max_payload=max_payload)
        self._transition = None

    @property
    def cargo(self):
        return self._cargo

    @property
    def payload(self):
        return self._cargo.payload

    @property
    def fullness(self):
        return self._cargo.fullness

    @property
    def is_empty(self):
        return self._cargo.is_empty

    @property
    def is_full(self):
        return self._cargo.is_full

    @property
    def free_space(self):
        return self._cargo.free_space

    @property
    def sprite_filename(self):
        return 'not-found.png'

    def game_step(self):
        if self._transition:
            if not self._transition.is_finished:
                self._transition.game_step()
            if self._transition.is_finished:
                if self._transition.cargo_to == self._cargo:
                    self.on_load_complete()
                if self._transition.cargo_from == self._cargo:
                    self.on_unload_complete()
                self._transition = None

        super(Unit, self).game_step()

    def move_at(self, target, speed=None):
        if self._move_target == target:
            return
        self._move_target = target
        super(Unit, self).move_at(target, speed=speed)

    def turn_to(self, target, speed=None):
        super(Unit, self).turn_to(target, speed=speed)

    def load_from(self, source):
        if self._cargo.free_space:
            self._transition = CargoTransition(cargo_from=source._cargo, cargo_to=self._cargo)

    def unload_to(self, target):
        if not self._cargo.is_empty:
            self._transition = CargoTransition(cargo_from=self._cargo, cargo_to=target._cargo)

    def on_load_complete(self):
        pass

    def on_unload_complete(self):
        pass


class Drone(Unit):
    rotate_mode = ROTATE_TURNING
    radius = 44
    auto_team = True
    layer = 2
    sleep_attrs = ('coord', 'vector', 'payload', 'gun_cooldown')

    class __DeathAnimation(object):
        def __init__(self, owner):
            self._owner = owner
            self.__sprites = ['teams/{}/blow_up_{}.png'.format(owner.team_number, i + 1) for i in range(3)]
            self.__sprites.append('teams/any_drone_explosion.png')
            self.__sprites.append('teams/{}/drone_crashed.png'.format(owner.team_number))
            self.__animation_speed = 10  # фреймов на спрайт
            self.__animation_frame = 0  # счетчик фреймов
            self.__current_sprite = self.__sprites[0]

        def sprite_filename(self):
            # последний в списке?
            if self.__sprites.index(self.__current_sprite) >= len(self.__sprites) - 1:
                return self.__current_sprite
            if self.__animation_frame < self.__animation_speed:
                self.__animation_frame += self._owner.scene.game_speed
            else:
                self.__animation_frame = 0
                # переключиться на следующий
                self.__current_sprite = self.__sprites[self.__sprites.index(self.__current_sprite) + 1]
            return self.__current_sprite

    def __init__(self, **kwargs):
        self._mothership = None
        self._gun = None
        if theme.DRONES_CAN_FIGHT:
            self._gun = PlasmaGun(self)
        self.__health = theme.DRONE_MAX_SHIELD
        super(Drone, self).__init__(max_payload=theme.MAX_DRONE_ELERIUM, **kwargs)

        self.__dead_flight_speed = theme.DRONE_DEAD_SPEED
        self.__angle_of_death = None
        self.__death_animaion = self.__DeathAnimation(self)

        self._sleep_state = None
        self._sleep_countdown = theme.SLEEP_COUNTDOWN

    @property
    def have_gun(self):
        return self._gun is not None

    @property
    def gun(self):
        return self._gun

    @property
    def health(self):
        return self.__health

    @property
    def gun_cooldown(self):
        if self._gun:
            return self._gun.cooldown

    @property
    def teammates(self):
        return [mate for mate in self.scene.drones
                if mate != self and mate.team == self.team and mate.is_alive]

    @property
    def sprite_filename(self):
        if self.is_alive:
            return 'teams/{}/drone.png'.format(self.team_number)
        return self.__death_animaion.sprite_filename()

    @property
    def meter_1(self):
        return min(self._cargo.fullness, 1.0)

    @property
    def meter_2(self):
        return float(self.__health) / theme.DRONE_MAX_SHIELD

    @property
    def is_alive(self):
        return self.__health > 0

    def damage_taken(self, damage=0):
        self.__health = max(self.__health - damage, 0)
        if self.__health <= 0:
            self.stop()

    def __heal_taken(self, healed_on=0):
        if self.__health > 0:
            self.__health = min(self.__health + healed_on, theme.DRONE_MAX_SHIELD)

    @property
    def mothership(self):
        if self._mothership is None:
            self._mothership = self.scene.get_mothership(self.team)
            assert self._mothership is not None
        return self._mothership

    def __dead_game_step(self):
        if self.__dead_flight_speed < 0.0:
            return
        if self.__angle_of_death is None:
            self.layer = 1
            self.__angle_of_death = random.randint(0, 359)

        x = self.__dead_flight_speed * math.sin(self.__angle_of_death)
        y = self.__dead_flight_speed * math.cos(self.__angle_of_death)
        self.coord.x = min(theme.FIELD_WIDTH - self.radius, max(self.radius, self.coord.x + x))
        self.coord.y = min(theme.FIELD_HEIGHT - self.radius, max(self.radius, self.coord.y + y))

        self.__dead_flight_speed -= theme.DRONE_DEAD_SPEED_DECELERATION * self.scene.game_speed
        super(Drone, self).game_step()

    def game_step(self):
        if not self.is_alive:
            self.__dead_game_step()
            return
        if self.mothership and self.mothership.is_alive and \
                self.distance_to(self.mothership.coord) < theme.MOTHERSHIP_HEALING_DISTANCE:
            self.__heal_taken(theme.MOTHERSHIP_HEALING_RATE)
        else:
            self.__heal_taken(theme.DRONE_SHIELD_RENEWAL_RATE)
        if self.have_gun:
            self.gun.game_step()
        super(Drone, self).game_step()
        self.update_sleep_state()
        if self.is_asleep():
            self.add_event(EventWakeUp())

    def move_at(self, target, speed=None):
        if not self.is_alive:
            return
        super(Drone, self).move_at(target, speed=theme.DRONE_SPEED)

    def turn_to(self, target, speed=None):
        if not self.is_alive:
            return
        super(Drone, self).turn_to(target, speed=theme.DRONE_TURN_SPEED)

    @property
    def my_mothership(self):
        return self.mothership

    @property
    def asteroids(self):
        return self.scene.asteroids

    def on_stop_at_target(self, target):
        for asteroid in self.asteroids:
            if asteroid.near(target):
                self.on_stop_at_asteroid(asteroid)
                return
        else:
            for ship in self.scene.motherships:
                if ship.near(target):
                    self.on_stop_at_mothership(ship)
                    return
        self.on_stop_at_point(target)

    def on_stop_at_point(self, target):
        pass

    def on_stop_at_asteroid(self, asteroid):
        pass

    def on_stop_at_mothership(self, mothership):
        pass

    def update_sleep_state(self):
        new_state = '-'.join([str(getattr(self, attr)) for attr in self.sleep_attrs])
        if self._sleep_state is None or self._sleep_state != new_state:
            self._sleep_state = new_state
            self._sleep_countdown = theme.SLEEP_COUNTDOWN
        elif self._sleep_countdown > 0:
            self._sleep_countdown -= 1
        else:
            self._sleep_countdown = theme.SLEEP_COUNTDOWN

    def is_asleep(self):
        return self._sleep_countdown <= 0

    def on_wake_up(self):
        self.info('Wake up, Neo!')


class Asteroid(Unit):
    rotate_mode = ROTATE_TURNING
    radius = 50
    selectable = False
    counter_attrs = dict(size=16, position=(0, 0), color=(255, 255, 255))
    layer = 1

    def __init__(self, elerium=None, **kwargs):
        if "direction" not in kwargs:
            kwargs["direction"] = random.randint(0, 360)
        self.__sprite_num = random.randint(1, 5)
        self._size = (elerium / theme.MIN_ASTEROID_ELERIUM) * .8
        super(Asteroid, self).__init__(payload=elerium, max_payload=elerium, **kwargs)

    @property
    def cargo(self):
        return self._cargo

    @property
    def sprite_filename(self):
        return 'asteroids/{}.png'.format(self.__sprite_num)

    @property
    def zoom(self):
        return .4 + self.cargo.fullness * .6 * self._size

    @property
    def is_alive(self):
        return False

    @property
    def counter(self):
        return self.cargo.payload if self._cargo is not None else 0

    def on_born(self):
        # super(Unit, self).on_born()
        self.turn_to(self.direction + 90, speed=theme.ASTEROID_ROTATION_SPEED)

    def on_stop(self):
        # super(Asteroid, self).on_stop()
        self.turn_to(self.direction + 90, speed=theme.ASTEROID_ROTATION_SPEED)

    def on_hearbeat(self):
        pass


class MotherShip(Unit):
    radius = 90
    selectable = False
    counter_attrs = dict(size=22, position=(75, 135), color=(255, 255, 255))
    layer = 1

    class __DeathAnimation(object):
        def __init__(self, owner):
            assert owner.team is not None
            self._owner = owner
            self.__sprites = ['motherships/{}/blow_up_{}.png'.format(owner.team_number, i + 1) for i in range(2)]
            self.__sprites.append('motherships/any_mothership_explosion.png')
            self.__sprites.append('motherships/{}/crashed.png'.format(owner.team_number))
            self.__animation_speed = 10  # фреймов на спрайт
            self.__animation_frame = 0  # счетчик фреймов
            self.__current_sprite = self.__sprites[0]

        def sprite_filename(self):
            # последний в списке?
            if self.__sprites.index(self.__current_sprite) >= len(self.__sprites) - 1:
                return self.__current_sprite
            if self.__animation_frame < self.__animation_speed:
                self.__animation_frame += self._owner.scene.game_speed
            else:
                self.__animation_frame = 0
                # переключиться на следующий
                self.__current_sprite = self.__sprites[self.__sprites.index(self.__current_sprite) + 1]
            return self.__current_sprite

    def __init__(self, max_payload=0, **kwargs):
        super(MotherShip, self).__init__(max_payload=max_payload, **kwargs)
        self.__health = theme.MOTHERSHIP_MAX_SHIELD
        # WARN: на момент создания материнского корабля не известен № комманды
        # поэтому создаем в момент первого обращения
        self.__death_animaion = None

    @property
    def sprite_filename(self):
        if self.is_alive:
            return 'teams/{}/mothership.png'.format(self.team_number)
        if self.__death_animaion is None:
            self.__death_animaion = self.__DeathAnimation(self)
        return self.__death_animaion.sprite_filename()

    @property
    def meter_1(self):
        return min(self._cargo.fullness, 1.0)

    @property
    def meter_2(self):
        return self.__health / theme.MOTHERSHIP_MAX_SHIELD

    @property
    def counter(self):
        return self.cargo.payload

    # TODO аналогичный метод для дрона - спрямить
    @property
    def health(self):
        return self.__health

    @property
    def is_alive(self):
        return self.__health > 0

    def damage_taken(self, damage=0):
        self.__health = max(self.__health - damage, 0)

    def __heal_taken(self, healed_on=0):
        if self.__health > 0:
            self.__health = min(self.__health + healed_on, theme.MOTHERSHIP_MAX_SHIELD)

    def game_step(self):
        self.__heal_taken(theme.MOTHERSHIP_SHIELD_RENEWAL_RATE)

    def on_hearbeat(self):
        pass

    def on_stop(self):
        pass



import os

import setuptools


README = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

with open('requirements.txt') as ff:
    install_requires = [line for line in ff.read().splitlines() if len(line) > 0]

setuptools.setup(
    name='astrobox',
    version='1.4.0',
    packages=setuptools.find_packages(),
    include_package_data=True,
    license='BSD License',
    description='The package allows you to create Astro Robo Game for programmers.',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/suguby/astrobox',
    author='Shandrinov Vadim',
    author_email='suguby@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    install_requires=install_requires,
)
